import click

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.tracked_command import TrackedCommand
from sym.flow.cli.models import ServiceType
from sym.flow.cli.models.service import BaseService
from sym.flow.cli.models.user import CSVMatcher


@click.command(
    cls=TrackedCommand,
    name="delete",
    short_help="Delete a User",
)
@click.argument("email", required=True, type=str)
@click.option(
    "--force",
    is_flag=True,
    help="Delete the user immediately",
)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def users_delete(options: GlobalOptions, email: str, force: bool) -> None:
    # Prompt user to confirm
    if not force:
        click.confirm(
            "WARNING: This is a destructive action. \nAre you sure you want to continue?",
            abort=True,
            default=False,
        )

    # External ID is not used, using service as a placeholder
    service_obj = BaseService(slug=ServiceType.SYM.type_name, external_id="cloud")
    matcher = CSVMatcher(service=service_obj, value=email)

    payload = {
        "users": [
            {
                "identity": {
                    "service_type": ServiceType.SYM.type_name,
                    "matcher": matcher.to_dict(),
                },
            }
        ]
    }

    options.sym_api.delete_user(payload)
    cli_output.success(f"Successfully deleted user {email}!")



import uuid
import webbrowser
from abc import ABC, abstractmethod
from http.server import HTTPServer
from typing import Optional, Tuple

import click
import pkce
import requests

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.errors import LoginError
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.login.handler import make_auth0_callback_handler
from sym.flow.cli.models import AuthToken, Organization


class LoginFlow(ABC):
    @abstractmethod
    def login(self, options: GlobalOptions, org: Organization) -> AuthToken:
        pass


class BrowserRedirectFlow(LoginFlow):
    """
    Issue an authorization code request by opening the browser to a particular
    URL. Start a server and wait for the user to successfully login. The browser will
    redirect a successful login back to the locally running server, with a request
    body containing an auth code. With the auth code, obtain an access token.

    Implements the auth code flow described here:
    https://www.altostra.com/blog/cli-authentication-with-auth0
    """

    def __init__(self, port: int):
        self.port: int = port
        self.redirect_url: str = f"http://localhost:{self.port}/callback"

        # Tuple of: url, state, code_verifier
        self.browser_login_params: Optional[Tuple[str, str, str]] = None

    def login(self, options: GlobalOptions, org: Organization):
        # Generate the Auth0 login URL
        (url, state, code_verifier) = self.gen_browser_login_params(options, org)

        # Print a copiable URL to the CLI
        styled_url = click.style(requests.utils.requote_uri(url), bold=True)  # type: ignore
        cli_output.info(
            f"Opening the login page in your browser. If this doesn't work, please visit the following URL:\n"
            f"{styled_url}\n"
        )

        # Start the HTTPServer that will listen for the callback after the user logs in.
        handler = make_auth0_callback_handler(options, self.redirect_url, state, code_verifier, org)
        with self._login_callback_server(handler) as httpd:
            # Open the URL in the user's browser.
            webbrowser.open(url)

            # Block and wait for the callback
            httpd.handle_request()

    def _login_callback_server(self, handler):
        """Starts an HTTPServer on the given port that will handle the callback from Auth0 after a user logs in on
        their browser.
        """
        try:
            return HTTPServer(("localhost", self.port), handler)
        except OSError as e:
            if e.errno in (48, 98):  # Address already in use
                raise LoginError(
                    f"Port {self.port} is already taken.",
                    "Try running `symflow login --port <port_number>` with an unused port.",
                )

    def gen_browser_login_params(
        self, options: GlobalOptions, org: Organization, force: bool = False
    ) -> Tuple[str, str, str]:
        """Generates the initial login URL that directs the user to log into Auth0."""

        if force or self.browser_login_params is None:
            state = str(uuid.uuid4())
            code_verifier = pkce.generate_code_verifier(length=128)
            code_challenge = pkce.get_code_challenge(code_verifier)
            query_params = "&".join(
                [
                    "response_type=code",
                    f"client_id={org.client_id}",
                    "code_challenge_method=S256",
                    f"code_challenge={code_challenge}",
                    f"redirect_uri={self.redirect_url}",
                    f"audience=https://api.symops.com",
                    f"state={state}",
                    "scope=login:cli",
                    "prompt=login",
                ]
            )
            url = f"{options.auth_url}/authorize?{query_params}"
            self.browser_login_params = (url, state, code_verifier)

        return self.browser_login_params



import click

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.helpers.config import Config
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.login.login_flow import BrowserRedirectFlow
from sym.flow.cli.helpers.tracked_command import TrackedCommand


# Note: This command deliberately does not define `cls=TrackedCommand`, as we track symflow logins when we
# hit the auth/login API (sym_api.verify_login)
@click.command(short_help="Log in to your Sym account")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.option(
    "--port",
    default=11001,
    help="Port to use for the local webserver.",
    show_default=True,
)
@click.option(
    "--org-id",
    prompt="Org ID",
    default=lambda: Config.get_org_slug(),
    help="The ID of the Organization you want to log in to.",
)
def login(options: GlobalOptions, port: int, org_id: str) -> None:
    """Log in to your Sym account to authenticate with the Sym API. This is required to enable privileged actions
    within your Organization such as authoring Sym Flows or installing the Slack App.

    \b
    Example:
        `symflow login --org-id S-0FNWNPZVI4`
    """
    if options.access_token:
        cli_output.fail("SYM_JWT is set, unset SYM_JWT to log in as a regular user")

    org = options.sym_api.get_organization_from_slug(org_id)

    # Save the organization to config so it will be autopopulated the next time `symflow login` is called.
    Config.set_org(org)

    flow = BrowserRedirectFlow(port)
    flow.login(options, org)


@click.command(cls=TrackedCommand, short_help="Log out of your Sym account")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def logout(options: GlobalOptions) -> None:
    if not Config.is_logged_in() and not options.access_token:
        cli_output.fail("You are already logged out!")

    if options.access_token:
        cli_output.fail("You are logged in via SYM_JWT, you must unset SYM_JWT manually")

    if Config.is_logged_in():
        Config.logout()
        cli_output.success("✔️  You successfully logged out!")



import os
from contextlib import contextmanager
from pathlib import Path

from .lock import read_lock, read_write_lock


class ThreadSafeFile:
    """A class that provides read/write access to a File protected by portalocker Locks
    https://portalocker.readthedocs.io/en/latest/index.html
    """

    def __init__(self, path: Path):
        # Make the path absolute, resolving all symlinks on the way and also normalizing it
        self.path = path.resolve()
        self.mkparents()

        # Create locking mechanisms for the given file
        self.read_lock = read_lock(self.path)
        self.update_lock = read_write_lock(self.path)

    def __str__(self):
        return str(self.path)

    @classmethod
    def xdg_config_home(cls) -> Path:
        """Returns the path to the .config directory.
        If the XDG_CONFIG_HOME envvar is set, uses that as the config directory.
        Otherwise, the .config directory defaults to ~/.config

        XDG Base Directory Specification Basics:
        https://specifications.freedesktop.org/basedir-spec/latest/ar01s02.html
        """
        try:
            return Path(os.environ["XDG_CONFIG_HOME"]).expanduser().resolve()
        except KeyError:
            return Path.home() / ".config"

    def mkparents(self):
        """Creates all the parent directories for the config file if they don't exist."""
        self.path.parent.mkdir(parents=True, exist_ok=True)

    @contextmanager
    def read(self):
        """Reads the given string from the file."""
        with self.read_lock as f:
            f.seek(0)  # Start cursor at the beginning of the file
            yield f

    @contextmanager
    def update(self):
        """Grabs a read-write lock on the config file for the duration of the context"""
        self.mkparents()
        with self.update_lock as f:
            f.seek(0)  # Start cursor at the beginning of the file
            yield f



import importlib.resources as pkg_resources
import os
import re
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path
from typing import Optional

import click

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.code_generation_templates import (  # import the *package* containing the tf files
    core,
)
from sym.flow.cli.errors import CliError
from sym.flow.cli.helpers.version import get_current_version

# Slugs are only allowed to have letters, numbers, and dashes. This regex matches
# all character that are NOT those.
slug_disallowed_pattern = re.compile(r"[^a-zA-Z\d-]+")


class CodeGenerationError(CliError):
    """Generic class for code generation errors."""


class MissingConfig(CodeGenerationError):
    def __init__(self, file_name: str, name: str) -> None:
        super().__init__(f"Could not generate {file_name}. A value for {name} is required, but was not given.")


class FlowGeneration(ABC):
    """Core logic for generating Flow configuration with `symflow generate`."""

    def __init__(self, flow_name: str) -> None:
        self.flow_name = flow_name
        self.generation_time = datetime.utcnow()
        self.symflow_version = str(get_current_version())

        # Subclasses will choose whether they need runtime.tf or not, so we may
        # not need aws_region. Set to None here so it's never undefined.
        self.aws_region: Optional[str] = None

    @property
    def flow_resource_name(self) -> str:
        """The name of the Flow used as a slug may have dashes, but used as a
        file name or Terraform resource should have underscores.
        """
        return self.flow_name.replace("-", "_").lower()

    @property
    def working_directory(self):
        return "."

    @abstractmethod
    def generate(self) -> None:
        """Generate all files required to configure this particular type of
        Flow.
        """

    def _get_aws_region(self) -> Optional[str]:
        """If a runtime.tf file would need to be generated, prompt the user for
        the AWS region to use.
        """

        if not Path("./runtime.tf").is_file():
            return click.prompt(
                click.style("What AWS region are your resources in?", bold=True), type=str, default="us-east-1"
            )

    def _generate_runtime_tf(self) -> None:
        """Generate runtime.tf if it does not already exist."""

        if not Path("./runtime.tf").is_file():
            if not self.aws_region:
                raise MissingConfig(file_name="runtime.tf", name="aws_region")

            runtime_tf = pkg_resources.read_text(core, "runtime.tf")

            with open("./runtime.tf", "w") as f:
                runtime_tf = runtime_tf.replace("SYM_TEMPLATE_VAR_SYMFLOW_VERSION", self.symflow_version)
                runtime_tf = runtime_tf.replace(
                    "SYM_TEMPLATE_VAR_GENERATION_TIME", self.generation_time.strftime("%Y-%m-%d %H:%M")
                )
                runtime_tf = runtime_tf.replace("SYM_TEMPLATE_VAR_AWS_REGION", self.aws_region)
                f.write(runtime_tf)

    def _generate_secrets_tf(self) -> None:
        """Generate secrets.tf if it does not already exist."""

        if not Path("./secrets.tf").is_file():
            secrets_tf = pkg_resources.read_text(core, "secrets.tf")

            with open("./secrets.tf", "w") as f:
                secrets_tf = secrets_tf.replace("SYM_TEMPLATE_VAR_SYMFLOW_VERSION", self.symflow_version)
                secrets_tf = secrets_tf.replace(
                    "SYM_TEMPLATE_VAR_GENERATION_TIME", self.generation_time.strftime("%Y-%m-%d %H:%M")
                )
                f.write(secrets_tf)

    def _create_impls_directory(self) -> None:
        """Create the impls directory to contain all Flow impls if it does not
        already exist.
        """

        if not Path("./impls").is_dir():
            os.makedirs("impls")


def get_valid_slug(label: str, slug_candidate: str) -> Optional[str]:
    """Interactively validate that a given string could be used as a slug and/or as a Terraform
    resource name. Outputs an error mesasge to the command line if not.

    Args:
        label: How to refer to the value the `slug_candidate` will be used for.
        slug_candidate: The string to validate.

    Returns:
        The `slug_candidate` if it's valid or the user confirmed a valid suggestion.
        None otherwise.
    """

    # Look for any disallowed characters in the given slug_candidate.
    if slug_disallowed_pattern.search(slug_candidate):
        # Replace any disallowed characters with dashes and remove any from the start or end.
        formatted_slug_candidate = slug_disallowed_pattern.sub("-", slug_candidate).strip("-")

        # It's possible there was nothing valid to work with, in which case just reject the name outright.
        if not formatted_slug_candidate:
            return None

        # Give the user a chance to accept the newly formatted slug. If they accept, great. If not, return None
        # so we can prompt again for a totally new name.
        if click.confirm(
            click.style(
                f'{label} can only contain letters, numbers, and dashes. Use "{formatted_slug_candidate}" instead?',
                fg="red",
            )
        ):
            return formatted_slug_candidate

        # Trigger a re-prompt for the slug
        return None

    # The slug will often be used in combination with some other identifying text (e.g. "SLUG-okta-group").
    # Resources like Targets have a slug limit of 55 characters (and -okta-group is 11), so 32 should give us a
    # little bit of wiggle room.
    if len(slug_candidate) > 32:
        cli_output.error(f"{label} cannot be longer than 32 characters. Please enter a shorter name.")

        # Trigger a re-prompt for the Flow name
        return None

    return slug_candidate



import importlib.resources as pkg_resources
from typing import Optional

import click

from sym.flow.cli.code_generation_templates import (  # import the *package* containing the tf files
    flows,
)

from .core import FlowGeneration


class OktaFlowGeneration(FlowGeneration):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.domain_name: str = click.prompt(
            click.style("What is your Okta domain?", bold=True), type=str, prompt_suffix=""
        )
        self.group_id: str = click.prompt(
            click.style("What is the Okta group ID you would like to manage with Sym?", bold=True),
            type=str,
            prompt_suffix="",
        )
        self.aws_region: Optional[str] = self._get_aws_region()

    @property
    def impl_filepath(self) -> str:
        return f"{self.working_directory}/impls/okta_{self.flow_resource_name}_impl.py"

    @property
    def flow_tf_filepath(self) -> str:
        return f"{self.working_directory}/okta_{self.flow_resource_name}.tf"

    def generate(self) -> None:
        """Generate the impl and Terraform files required to configure an Okta Flow."""
        # Generate any core requirements that don't already exist in this directory.
        self._generate_runtime_tf()
        self._generate_secrets_tf()
        self._create_impls_directory()

        # Generate the Okta-specific files.
        with open(self.impl_filepath, "w") as f:
            # Note: The impl file is stored as a `.txt` resource because PyOxidizer (the tool used to package symflow CLI)
            # Does NOT support reading `.py` files with `importlib.resources`
            # https://github.com/indygreg/PyOxidizer/issues/237
            #
            # However, we don't care about reading the source code, we simply need to pull the text file and write it
            # to the filesystem with a `.py` extension. As a workaround, we have stored `impl.py` as `impl.txt` in the
            # code_generation_templates.flows package so that we can read it with importlib.resources.
            impl_txt = pkg_resources.read_text(flows, "impl.txt")
            f.write(impl_txt)

        with open(self.flow_tf_filepath, "w") as f:
            okta_tf = pkg_resources.read_text(flows, "okta.tf")

            okta_tf = okta_tf.replace("SYM_TEMPLATE_VAR_SYMFLOW_VERSION", self.symflow_version)
            okta_tf = okta_tf.replace(
                "SYM_TEMPLATE_VAR_GENERATION_TIME", self.generation_time.strftime("%Y-%m-%d %H:%M")
            )
            okta_tf = okta_tf.replace("SYM_TEMPLATE_VAR_FLOW_RESOURCE_NAME", self.flow_resource_name)
            okta_tf = okta_tf.replace("SYM_TEMPLATE_VAR_FLOW_NAME", self.flow_name)
            okta_tf = okta_tf.replace("SYM_TEMPLATE_VAR_OKTA_DOMAIN", self.domain_name)
            okta_tf = okta_tf.replace("SYM_TEMPLATE_VAR_OKTA_GROUP_ID", self.group_id)
            f.write(okta_tf)



import sys
import urllib.parse as parse
from http.server import BaseHTTPRequestHandler

import requests

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.errors import SymAPIRequestError
from sym.flow.cli.helpers.config import Config
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.jwt import JWT
from sym.flow.cli.models import AuthToken, Organization


def make_auth0_callback_handler(
    options: GlobalOptions, redirect_uri: str, state: str, code_verifier: str, organization: Organization
):
    """A Method that returns an HTTPRequestHandler class that can be passed into an HTTPServer instantiation.

    Handles a GET request callback from Auth0 after a user logs in:
    - Redirects the user to /cli/error if any errors occur (e.g. Invalid state parameter)
    - Exchanges the auth_code for an access token
    - Calls Queuer's VerifyLogin endpoint to validate that the user is a valid user in the organization.

    Note: The class has to be defined within this method because we can't change the shape of the __init__ method, as
    the handler is instantiated by the HTTPServer code upon each request (We only pass in the class type of the handler
    when we start the HTTPServer, not an instance of the handler). However, these values are needed to exchange the
    auth code and verify the login, so the class definition will reference the values passed in from this wrapper method
    and the wrapper will return the class reference that can be passed into HTTPServer.

    Args:
        options: A GlobalOptions instance that can be used to call the Queuer API
        redirect_uri: The redirect_uri parameter sent to Auth0 in the initial login URL
        state: The state parameter sent to Auth0 in the initial login URL
        code_verifier: The plaintext string used to generate the code_challenge sent to Auth0 in the initial login URL
        organization: The User's Organization
    """

    class Auth0HTTPRequestHandler(BaseHTTPRequestHandler):
        """A HTTPRequestHandler that handles the OAuth2.0 callback from Auth0 when a user logs in.
        See: https://auth0.com/docs/get-started/authentication-and-authorization-flow/authorization-code-flow-with-proof-key-for-code-exchange-pkce
        """

        def log_message(self, format, *args):
            """BaseHTTPRequestHandler will log requests and other messages to stderr. Override log_message to prevent
            any unnecessary prints to the CLI.
            """

        def do_GET(self):
            """Handles the callback from Auth0 after a user logs in on their browser."""
            parsed = parse.urlparse(self.path)

            # Checks if Auth0 returned any error messages and exits with an error if so.
            # Validates that both 'code' and 'state' parameters exists. Also validates that the state parameter
            # matches the state parameter generated at the beginning of the login process.
            code = self._validate_callback_params(parsed.query)

            # We're guaranteed to have an auth code at this point that we can exchange for an access token.
            auth_token = self._get_token_from_code(code)

            # Set the access token in GlobalOptions so that we can make API calls to Queuer
            options.set_access_token(auth_token.access_token)

            # Call Queuer's ValidateLogin API to verify that the user is a valid user in the Organization
            # (At this point we only know that they have a valid Auth0 login.)
            self._validate_login(auth_token)

            # Print success to the CLI, and redirect the browser to the Login Success page.
            cli_output.success("Login succeeded")
            self.send_response(301)
            self.send_header("Location", "https://static.symops.com/cli/login")
            self.end_headers()

        def _validate_callback_params(self, query: str):
            """
            Validates the callback query parameters. If:
                - An error_description exists
                - The `code` parameter is missing
                - The `state` parameter is missing, or doesn't match the state parameter we generated
            then the error is reported in both the browser and CLI. Then exits the CLI with exit code 1.

            Args:
                query: The query parameters passed into the GET request.

            Returns:
                The auth code from Auth0 (i.e. the 'code' parameter)
            """
            qs = parse.parse_qs(query)

            error_message = None
            if qs.get("error_description"):
                error_message = qs["error_description"][0]
            elif not qs.get("code"):
                error_message = "Missing code in query"
            elif not qs.get("state"):
                error_message = "Missing state in query"
            elif not qs["state"][0] == state:
                options.dprint(f"Invalid state. Found: {qs['state'][0]}, expected: {state}.")
                error_message = "Invalid state"

            # Redirect the browser with the error page and print the error to the CLI
            if error_message:
                self._exit_with_error(error_message)

            return qs["code"][0]

        def _exit_with_error(self, error_message: str):
            """Redirects the browser to the /cli/error page with the given error message, prints it to the CLI, then
            exits with an error code.
            """

            query = parse.urlencode({"message": error_message})
            self.send_response(301)
            self.send_header("Location", f"https://static.symops.com/cli/error?{query}")
            self.end_headers()

            if "complete sign up" in error_message:
                # If we know the problem is that the user just signed up (and therefore their email is not validated),
                # give them an info message telling them to verify their email.
                cli_output.info(error_message)
                sys.exit(0)
            else:
                # Otherwise, raise the error message that we got from the HTTP response.
                cli_output.fail(error_message)

        def _get_token_from_code(self, auth_code: str) -> AuthToken:
            """With the provided auth code and the initial code verifier, exchanges the auth code for an Auth0
            access token.
            See: https://auth0.com/docs/get-started/authentication-and-authorization-flow/authorization-code-flow-with-proof-key-for-code-exchange-pkce
            """
            data = {
                "grant_type": "authorization_code",
                "client_id": organization.client_id,
                "code_verifier": code_verifier,
                "code": auth_code,
                "redirect_uri": redirect_uri,
            }
            headers = {"content-type": "application/x-www-form-urlencoded"}
            url = f"{options.auth_url}/oauth/token"
            r = requests.post(url, headers=headers, data=data)
            if not r.ok:
                self._exit_with_error(
                    f"An unexpected error occurred while logging in, please try again later. {r.text}"
                )

            auth_token = AuthToken.parse_obj(r.json())
            return auth_token

        def _validate_login(self, auth_token: AuthToken):
            """Makes an API Call to Queuer to verify that the user that just logged in is a valid admin in the
            Organization. If not, then reports that error to the browser/CLI and exits.

            Args:
                auth_token: The AuthToken object received from a successful Auth0 login
            """

            jwt = JWT.from_access_token(auth_token.access_token, auth_url=options.auth_url)

            # Verify user exists in the organization
            try:
                # Check the User's session with the Sym API. This will validate that the JWT is still valid,
                # and matches the email address that the User is logged in with.
                user_data = options.sym_api.verify_login(segment_track=True)
            except SymAPIRequestError:
                self._exit_with_error("A server error has occurred, please try again later.")

            if user_data.get("error") and user_data.get("message"):
                self._exit_with_error(user_data["message"])
            elif user_data.get("error"):
                # All Sym API errors should have a message field, but just in case...
                self._exit_with_error(f"{jwt.email} is not a valid user in the Organization {organization.slug}")

            # We successfully validated the user. Save the access token to the symflow config
            Config.store_login_config(jwt.email, organization, auth_token)

    return Auth0HTTPRequestHandler



from typing import Dict, List, Optional
from uuid import UUID

from pydantic import parse_obj_as

from sym.flow.cli.errors import (
    SymAPIUnknownError,
    UnknownBotError,
    UnknownOrgError,
    UnknownUserError,
)
from sym.flow.cli.helpers.config import Config
from sym.flow.cli.helpers.constants import SegmentTrackStatus
from sym.flow.cli.helpers.rest import SymRESTClient
from sym.flow.cli.helpers.utils import filter_dict
from sym.flow.cli.models import Organization
from sym.flow.cli.models.resource import ResourceType, TerraformResource
from sym.flow.cli.models.service import Service
from sym.flow.cli.models.service_type import ServiceType
from sym.flow.cli.models.token import SymToken
from sym.flow.cli.models.user import User
from sym.flow.cli.models.user_type import UserType


class SymAPI:
    def __init__(self, url: str, access_token: Optional[str] = None):
        self.rest = SymRESTClient(url=url, access_token=access_token or Config.get_access_token())

    def set_access_token(self, access_token: str):
        self.rest.access_token = access_token

    def get_current_organization(self) -> Organization:
        """Get the organization for the currently authenticated user."""

        # There should always be exactly one organization, since the API returns specifically
        # the single organization associated with the JWT.
        org_data = self.rest.get("organizations").json()[0]
        return Organization.parse_obj(org_data)

    def get_organization_from_slug(self, org_slug: str) -> Organization:
        """Retrieves Organization data (e.g. Auth0 Client ID) from the Sym API given
        an org slug (publicly referred to as Org ID).
        """

        org_data = self.rest.get(f"auth/org/{org_slug}", force_auth=False).json()
        org = Organization.parse_obj(org_data)

        # auth/org/<slug> should always return client_id, but other endpoints don't, so it's
        # not enforced on the model. i.e. this should never happen, but check JUST in case.
        if not org.client_id:
            raise UnknownOrgError(org_id=org_slug)

        return org

    def verify_login(self, email: Optional[str] = None, segment_track: Optional[bool] = False) -> dict:
        """Calls the Sym API to validate that the User's current credentials are valid.

        Returns:
            A dictionary containing User data if the credentials were validated, or error information
            if they were not.
        """
        response = self.rest.get(
            "auth/login", filter_dict({"email": email, "segment_track": segment_track}), validate=False
        )

        return response.json()

    def get_integrations(self) -> List[dict]:
        """Retrieve all Sym Integrations accessible to the currently
        authenticated user.
        """
        return self.rest.get("entities/integrations").json()

    def get_users(self, query_params: Optional[dict] = None) -> List[User]:
        """Retrieve all Sym Users accessible to the currently
        authenticated user.
        """

        response = self.rest.get("users", query_params).json()
        return parse_obj_as(List[User], response["users"])

    def get_user(self, email: str) -> User:
        """Given an email, get the whole user object and their identities"""
        users = self.get_users(query_params={"email": email})
        if len(users) != 1:
            raise UnknownUserError(email=email)

        return users[0]

    def get_user_by_id(self, user_id: UUID) -> Optional[User]:
        """Given a User UUID, get the whole user object and their identities"""
        users = self.get_users({"id": user_id})
        if users:
            return users[0]

        return None

    def get_bot(self, username: str) -> User:
        """Given a bot username, get the whole bot object and their identities"""
        bots = self.get_users({"type": UserType.BOT})
        bot = next((b for b in bots if b.sym_identifier == username), None)

        if not bot:
            raise UnknownBotError(username=username)

        return bot

    def update_users(self, payload: dict) -> List[dict]:
        return self.rest.patch("users", payload).json()

    def set_user_role(self, user_id: str, role: str) -> None:
        """Set the role of the user with the given user_id to the given role."""
        self.rest.post(f"users/{user_id}/set-role", data={"role": role})

    def delete_user(self, payload: dict) -> List[dict]:
        return self.rest.post("users/delete", payload).json()

    def delete_identities(self, payload: dict) -> List[dict]:
        return self.rest.post("users/identities/delete", payload).json()

    def get_slack_install_url(self, service_id: str) -> str:
        """Get the URL that starts the Slack App installation flow.

        Returns:
            The URL to install the Slack App and the Sym Request ID used to
            retrieve the URL as a string.
        """

        return self.rest.get("services/slack/link", {"service_id": service_id}).json()["url"]

    def uninstall_slack(self, service_id: str) -> None:
        """Make a request to the Sym API to uninstall the Slack App.

        Raises SymAPIUnknownError from handle_response on failure. Otherwise,
        assume success.
        """

        self.rest.get("services/slack/uninstall", {"token": service_id}, validate=False)

    def get_services(self, service_types: List[str] = ServiceType.all_names()) -> List[Service]:
        """Retrieve services in service_types registered to the currently
        authenticated user's organization.

        service_types defaults to all recognized service types
        """
        response = self.rest.get("services").json()
        services = parse_obj_as(List[Service], response["services"])

        return [s for s in services if s.service_type in service_types]

    def get_service(self, service_type: str, external_id: str) -> Service:
        """Retrieve a Service by the given service_type and external_id.
        Expects that the response will have exactly one Service
        """
        response = self.rest.get("services", {"service_type": service_type, "external_id": external_id}).json()
        services = parse_obj_as(List[Service], response["services"])
        if (num_services := len(services)) != 1:
            raise SymAPIUnknownError(
                response_code=400,
                request_id=self.rest.last_request_id,
                message=f"Expected 1 service, but found {num_services} instead",
            )

        return services[0]

    def update_service(self, service_type_name: str, external_id: str, label: str):
        """Updates a Service from the currently authenticated user's organization"""
        return self.rest.patch(
            "services",
            {"service_type": service_type_name, "external_id": external_id, "label": label},
        ).json()

    def delete_service(self, service_type_name: str, external_id: str):
        """Deletes a Service from the currently authenticated user's organization"""
        return self.rest.post(
            "services/delete",
            {"service_type": service_type_name, "external_id": external_id},
        ).json()

    def get_service_references(self, service_id: str) -> Dict[str, List[str]]:
        """Gets all objects referencing a service, i.e. Identities/Integrations
        if any exist, a service cannot be deleted"""
        return self.rest.get(f"service/{service_id}/references").json()["references"]

    def create_token(self, username: str, expiry_in_seconds: int, label: Optional[str] = None) -> str:
        """Creates a new token in the organization for the given bot-user and expiry"""
        response = self.rest.post(
            "tokens",
            filter_dict({"username": username, "expiry": expiry_in_seconds, "label": label}),
        ).json()

        return response["access_token"]

    def revoke_token(self, jti: str):
        return self.rest.post(
            "tokens/delete",
            {"identifier": jti},
        ).json()

    def get_tokens(self) -> List[SymToken]:
        response = self.rest.get(
            "tokens",
        ).json()
        return parse_obj_as(List[SymToken], response["tokens"])

    def get_resources(self, resource_type: ResourceType) -> List[TerraformResource]:
        """Retrieve all of an Organization's Terraform resources of a particular type."""

        response = self.rest.get(f"entities/{resource_type}").json()
        return parse_obj_as(List[TerraformResource], response)

    def add_domain(self, domain: str) -> None:
        """Add a domain to the organization's list."""
        org_id = self.get_current_organization().id
        self.rest.post(f"organizations/{org_id}/add-domain", {"domain": domain})

    def remove_domain(self, domain: str) -> None:
        """Remove a domain from the organization's list."""
        org_id = self.get_current_organization().id
        self.rest.post(f"organizations/{org_id}/remove-domain", {"domain": domain})

    def segment_track(
        self,
        subcommand: str,
        status: SegmentTrackStatus,
        error: Optional[str] = None,
    ):
        """
        Calls Queuer to track the symflow command in Segment.

        Args:
            subcommand: The full symflow subcommand (e.g. `users set-role`)
            status: One of: "invoked", "success", or "error"
            error: The error message, if status == error
        """
        try:
            self.rest.post(
                "track/symflow",
                {"subcommand": subcommand, "status": status, "error": error},
                validate=False,
            )
        except Exception:
            # We don't want to throw any errors on tracking.
            # Rest API Errors will already be captured in Sentry in rest._request(), so we don't need an additional
            # capture_exception here.
            pass



import os
from pathlib import Path

from portalocker import Lock
from portalocker.constants import LOCK_EX, LOCK_NB, LOCK_SH

""" 
    For a refresher on file-opening modes:
     `r`   Open text file for reading.  The stream is positioned at the
             beginning of the file.
    
     `r+`  Open for reading and writing.  The stream is positioned at the
             beginning of the file.
"""


def read_lock(path: Path):
    """
    Returns a Portalocker Read Lock for the file specified at `path`.
    Treat this method like the `open(filename, 'r')` Python method.

    Args:
        path: The path to the file (e.g. Path("~/.config/symflow/default/config.yml")
    """

    return Lock(
        str(path),
        # read-only mode
        mode="r",
        # LOCK_SH = Place a shared lock. More than one process may hold a shared lock for a given file at a given time
        # LOCK_NB = Acquire the lock in a non-blocking fashion
        flags=LOCK_SH | LOCK_NB,
        # Create the file on open if it doesn't exist.
        opener=lambda path, flags: os.open(path, flags | os.O_CREAT),
    )


def read_write_lock(path: Path):
    """
    Returns a Portalocker Read-Write Lock (r+) for the file specified at `path`.
    If the file doesn't exist yet, creates the file first.
    Treat this method like the `open(filename, 'r+')` Python method.

    This is a "spinlock" (https://en.wikipedia.org/wiki/Spinlock)

    Args:
        path:The path to the file (e.g. Path("~/.config/symflow/default/config.yml")
    """

    return Lock(
        str(path),
        # Read/write mode
        mode="r+",
        # LOCK_EX = exclusive lock. Only one process may hold an exclusive lock at a time
        # LOCK_NB = Acquire the lock in a non-blocking fashion
        flags=LOCK_EX | LOCK_NB,
        # Create the file on open if it doesn't exist.
        opener=lambda path, flags: os.open(path, flags | os.O_CREAT),
    )



from sym.flow.cli.helpers.api_operations import (
    Operation,
    OperationHelper,
    OperationSets,
    OperationType,
)
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.models import User
from sym.flow.cli.models.service import SYM_CLOUD_KEY


def update_full_name(
    options: GlobalOptions,
    user_to_update: User,
    first_and_middle_name: str,
    last_name: str,
):
    """Given a user, executes the operation helper for creating/updating the first (including middle) and last name of user in Sym identity"""

    sym_identity = user_to_update.get_identity_from_key(SYM_CLOUD_KEY)
    sym_identity.profile["first_name"] = first_and_middle_name
    sym_identity.profile["last_name"] = last_name

    operations = OperationSets(
        # Don't need to send in all the existing identities,
        # just the identity to update and the Sym identity to identify the user
        update_user_ops=[
            Operation(
                operation_type=OperationType.update_user,
                original_value=None,
                new_value=User(id=user_to_update.id, identities=[sym_identity]),
            )
        ],
        delete_identities_ops=[],
        delete_user_ops=[],
    )

    operation_helper = OperationHelper(options, operations)
    operation_helper.handle_update_users()



import importlib.resources as pkg_resources
import os
from datetime import datetime
from typing import Optional, cast

import click

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.code_generation_templates import (
    core,  # import the *package* containing the tf files
)
from sym.flow.cli.errors import NotLoggedInError
from sym.flow.cli.helpers.code_generation.core import get_valid_slug
from sym.flow.cli.helpers.config import Config
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.tracked_command import TrackedCommand
from sym.flow.cli.helpers.version import get_current_version


@click.command(cls=TrackedCommand, short_help="Generate starter Terraform code")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.option(
    "--slack-workspace-id", "--workspace-id", help="The ID of the Slack workspace where you installed the Sym App."
)
@click.option(
    "--environment",
    help="A name describing the environment where your Flows will be deployed (e.g. prod, staging, sandbox).",
)
def init(options: GlobalOptions, slack_workspace_id: Optional[str] = None, environment: Optional[str] = None) -> None:
    """Generates the common Terraform configuration required to start building any Sym Flow.

    After running this command, run `symflow generate` inside the created directory to generate
    the code for your first Sym Flow.
    """
    if not Config.is_logged_in() or not (org := Config.get_org()):
        raise NotLoggedInError()

    cli_output.info(
        "\nWelcome to Sym! This command will generate a new directory containing the Terraform files needed "
        "to start building your first Sym Flow.\n",
    )

    while True:
        init_dir = click.prompt(
            click.style("What should we name the directory?", bold=True), type=str, prompt_suffix=""
        )

        # Check if a file/directory already exists at the given path
        if os.path.exists(init_dir):
            cli_output.error(
                f"Looks like the directory '{init_dir}' already exists! Please provide the name for a new directory."
            )
        else:
            break

    if not slack_workspace_id:
        slack_workspace_id = cast(
            str,
            click.prompt(
                f"{click.style('Slack Workspace ID', bold=True)} {click.style('(You can find this in Slack by running /sym whoami)', dim=True, italic=True)}",
                type=str,
            ),
        )

    if environment:
        environment = get_valid_slug(label="Environment Name", slug_candidate=environment)

    while not environment:
        environment_candidate = cast(
            str,
            click.prompt(
                f"{click.style('Environment Name', bold=True)} {click.style('This is where your Sym Flows will be deployed (e.g. prod, staging, sandbox)', dim=True, italic=True)}",
                type=str,
                default="prod",
            ),
        )

        environment = get_valid_slug(label="Environment Name", slug_candidate=environment_candidate)

    # Create the directory
    os.makedirs(init_dir)

    # These files don't need any values substituted in them, and can just be copied over as-is.
    static_files = ["versions.tf", "README.md"]

    for output_file in static_files:
        template = pkg_resources.read_text(core, output_file)
        with open(f"{init_dir}/{output_file}", "w") as f:
            f.write(template)

    # Create the environment.tf file with Org ID, environment name, and workspace ID filled in
    environment_tf = pkg_resources.read_text(core, "environment.tf")
    with open(f"{init_dir}/environment.tf", "w") as f:
        environment_tf = environment_tf.replace("SYM_TEMPLATE_VAR_SYMFLOW_VERSION", str(get_current_version()))
        environment_tf = environment_tf.replace(
            "SYM_TEMPLATE_VAR_GENERATION_TIME", datetime.utcnow().strftime("%Y-%m-%d %H:%M")
        )
        environment_tf = environment_tf.replace("SYM_TEMPLATE_VAR_ORG_ID", org.slug)
        environment_tf = environment_tf.replace("SYM_TEMPLATE_VAR_SLACK_WORKSPACE_ID", slack_workspace_id.upper())
        environment_tf = environment_tf.replace("SYM_TEMPLATE_VAR_ENVIRONMENT_NAME", environment)

        f.write(environment_tf)

    cli_output.info(
        "\nSuccessfully generated your Sym Terraform configuration! Run the following to add your first Flow:"
    )
    cli_output.actionable(f"cd {init_dir} && symflow generate")



import click

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.errors import SymAPIRequestError
from sym.flow.cli.helpers.api import SymAPI
from sym.flow.cli.helpers.config import Config
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.tracked_command import TrackedCommand


@click.command(cls=TrackedCommand, short_help="Check your stored auth token")
@click.make_pass_decorator(GlobalOptions, ensure=True)
def status(options: GlobalOptions) -> None:
    """Check if you have an existing session that is still valid."""
    if options.access_token:
        check_status_by_sym_jwt(options.sym_api)
    else:
        check_status_by_config(options.sym_api)


def check_status_by_sym_jwt(api: SymAPI):
    """Validate that the environment variable SYM_JWT contains a valid JWT."""
    try:
        user_data = api.verify_login()
    except SymAPIRequestError:
        cli_output.fail("A server error has occurred, Please try again later.")

    if user_data.get("error") and user_data.get("message"):
        if user_data.get("code") == "Authentication:INVALID_JWT":
            # This specific error message is very generic, and the Bot Token has probably just expired.
            cli_output.fail(
                "Status check failed!",
                f"  {user_data['message']} Use `symflow tokens list` to make sure your token is still valid.",
            )
        else:
            cli_output.fail("Status check failed!", f"  {user_data['message']}")
    elif user_data.get("error"):
        # All Sym API errors should have a message field, but just in case...
        cli_output.fail(
            "Status check failed!",
            "   Please check the Sym documentation on Bot Users and Tokens at https://docs.symops.com/docs/using-bot-tokens",
        )

    org = user_data["organization"]
    username = user_data["username"]

    cli_output.success("✔️  Status check succeeded!")
    click.echo(
        f"   You are logged in to {click.style(org, bold=True)} as {click.style(username, bold=True)} via SYM_JWT."
    )


def check_status_by_config(api: SymAPI):
    """Validate that the User's auth token stored in config.yml is a valid JWT."""
    if not Config.is_logged_in():
        # If the User doesn't even have an auth token set, they are definitely not logged in.
        cli_output.fail("You are not currently logged in", "   Try running `symflow login`")

    email = Config.get_email()

    try:
        # Check the User's session with the Sym API. This will validate that the JWT is still valid,
        # and matches the email address that the User is logged in with.
        user_data = api.verify_login(email)
    except SymAPIRequestError:
        cli_output.fail("A server error has occurred, Please try again later.")

    if user_data.get("error") and user_data.get("message"):
        if user_data.get("code") == "Authentication:INVALID_JWT":
            # This specific error message is very generic, and the User probably just needs to log in again.
            cli_output.fail("Status check failed!", f"  {user_data['message']} Try running `symflow login`.")
        else:
            cli_output.fail("Status check failed!", f"  {user_data['message']}")
    elif user_data.get("error"):
        # All Sym API errors should have a message field, but just in case...
        cli_output.fail("Status check failed!", "   Try running `symflow login`.")

    org = user_data["organization"]
    username = user_data["username"]
    role = user_data["role"]

    cli_output.success("✔️  Status check succeeded!")
    # Not using cli_output here due to the complexity of the styling
    click.echo(
        f"   You are logged in to {click.style(org, bold=True)} as {click.style(username, bold=True)} (role: {click.style(role, bold=True)})."
    )



"""CLI Helpers"""

import sys
from typing import Optional

import click


def success(message: str, newline=True):
    click.secho(message, fg="green", nl=newline)


def warn(message: str):
    click.secho(message, fg="yellow")


def info(message: str, bold=False):
    click.secho(message, bold=bold)


def actionable(message: str):
    """Print a message that we want the user to act on."""
    click.secho(message, bold=True, fg="cyan")


def error(message: str, color="red"):
    click.secho(message, fg=color, err=True)


# This function is for errors that are followed by program exit
def fail(
    message: str = "Something went wrong",
    hint: Optional[str] = "Please check the Sym documentation at https://docs.symops.com/",
):
    """Raise a usage error with a useful message"""
    click.secho(f"✖ {message}", fg="red", bold=True, err=True)

    if hint:
        click.secho(f"{hint}", fg="cyan")

    sys.exit(1)



import click

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.helpers.tracked_command import TrackedCommand
from sym.flow.cli.version import __version__


@click.command(cls=TrackedCommand, short_help="Print the version")
def version() -> None:
    cli_output.info(__version__)



from pathlib import Path
from typing import Optional

import click

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.errors import NotLoggedInError
from sym.flow.cli.helpers.code_generation.approval_only import ApprovalFlowGeneration
from sym.flow.cli.helpers.code_generation.core import get_valid_slug
from sym.flow.cli.helpers.code_generation.okta import OktaFlowGeneration
from sym.flow.cli.helpers.config import Config
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.tracked_command import TrackedCommand
from sym.flow.cli.helpers.utils import get_or_prompt

FLOW_TYPE_OPTIONS = ["okta", "approval-only"]


@click.command(cls=TrackedCommand, short_help="Generate Terraform to configure a Sym Flow.")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.option(
    "--type",
    "type_",
    type=click.Choice(FLOW_TYPE_OPTIONS, case_sensitive=False),
    help="What type of Sym Flow to create.",
)
@click.option(
    "--flow-name",
    help="A unique name for this Sym Flow.",
)
def generate(options: GlobalOptions, type_: str, flow_name: Optional[str] = None) -> None:
    """Generates all Terraform required to configure a Sym Flow.

    This command should be run inside the directory created by `symflow init`.
    """
    if not Config.is_logged_in() or not Config.get_org():
        raise NotLoggedInError()

    # If environment.tf and versions.tf are not both present, we can relatively safely assume that
    # the user isn't inside a directory created by `symflow init` because it would have generated them.
    if not Path("./environment.tf").is_file() or not Path("./versions.tf").is_file():
        cli_output.fail(
            "This command must be run inside a directory created by `symflow init`. Please run that command first.",
            hint="Hint: The current directory must have `environment.tf` and `versions.tf` files.",
        )

    cli_output.info(
        "\nWelcome to Sym! This command will generate new files in the current directory required to configure a Sym Flow.\n",
    )

    # Prompt for any values that weren't already passed in.
    type_ = get_or_prompt(
        type_,
        click.style("Please select the type of Flow you would like to create", bold=True),
        FLOW_TYPE_OPTIONS,
    )

    # If the user used --flow-name to pass in a name, we need to do the same validation we'd do
    # if we were prompting for it.
    if flow_name:
        # This may set the flow_name back to None, if we need to re-prompt.
        flow_name = get_valid_slug(label="Flow name", slug_candidate=flow_name)

    while not flow_name:
        flow_name_candidate = click.prompt(
            f"{click.style('What would you like to name this Flow?', bold=True)} {click.style('This name should be unique', dim=True, italic=True)}",
            type=str,
        )

        flow_name = get_valid_slug(label="Flow name", slug_candidate=flow_name_candidate)

    cli_output.info(f'\nGenerating a Sym Flow of type "{type_}" named "{flow_name}"...\n')

    if type_ == "okta":
        OktaFlowGeneration(flow_name=flow_name).generate()
    elif type_ == "approval-only":
        ApprovalFlowGeneration(flow_name=flow_name).generate()
    else:
        # This should never happen, because the user chooses from a given list.
        # If we get here, we've done something wrong.
        cli_output.fail(
            f'Unknown Flow type: "{type_}". Please use `symflow generate --help` to see the list of available Flow types.'
        )

    cli_output.info(
        f"\nSuccessfully generated your Terraform configuration! Run the following to check the configuration:"
    )
    cli_output.actionable(f"terraform init && terraform plan")
    cli_output.info("\nWhen you are ready to apply your configuration, run the following:")
    cli_output.actionable("terraform apply")



from typing import List, Optional, Tuple

import click

from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.profile_operations import update_full_name
from sym.flow.cli.helpers.tracked_command import TrackedCommand


@click.command(cls=TrackedCommand, name="update-name", short_help="Update a User's name")
@click.argument("email", required=True, type=str)
@click.argument("full_name", required=True, type=str, nargs=-1)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def update_name(
    options: GlobalOptions,
    email: str,
    full_name: Tuple[str],
) -> None:
    """For an existing Sym User, update their full name in their Sym identity

    \b
    Example:
        1. `symflow users update-name beyonce@symops.io Beyonce`
        2. `symflow users update-name anupya@symops.io Anupya Pamidimukkala`
        2. `symflow users update-name neil@symops.io Neil Patrick Harris Martin Sheen`
    """

    api = options.sym_api
    user_to_update = api.get_user(email)
    first_and_middle_name, last_name = split_name(full_name)

    update_full_name(options, user_to_update, first_and_middle_name, last_name)


def split_name(name: Tuple[str]) -> Tuple[List[str], Optional[str]]:
    """Splits a given name into first and last name. If there are more
    than two words, all but the last word are combined into the first name.
    """

    if len(name) == 1:  # mononyms
        return name[0], ""
    else:
        return " ".join(name[:-1]), name[-1]



import os
from functools import wraps

import click
from sentry_sdk import Scope
from sentry_sdk.api import configure_scope
from sentry_sdk.hub import init as sentry_init

from sym.flow.cli.helpers.config import Config
from sym.flow.cli.helpers.utils import filter_dict
from sym.flow.cli.helpers.version import is_local


def getuser():
    """Get the username from the environment or password database.

    First try various environment variables, then the password
    database.  This works on Windows as long as USERNAME is set.

    """

    for name in ("LOGNAME", "USER", "LNAME", "USERNAME"):
        user = os.environ.get(name)
        if user:
            return user

    # If this fails, the exception will "explain" why
    import pwd

    return pwd.getpwuid(os.getuid())[0]


def set_scope_context_os(scope: Scope) -> None:
    """Save the user's operating system details in the Sentry context."""

    u = os.uname()
    uname_str = f"{u.sysname} {u.nodename} {u.release} {u.version} {u.machine}"
    scope.set_context(
        "os",
        {
            "name": u.sysname,
            "version": u.release,
            "build": u.version,
            "kernel_version": uname_str,
        },
    )


def setup_sentry(**kwargs):
    """Returns a decorator that will configure Sentry before the command is invoked.
    This decorator should be set on the base `symflow` command.

    - Sets the environment to `pypi`
    - Initializes the Sentry SDK
    - Sets Sentry tags identifying the user and their org
    - Sets the OS details in the Sentry context
    """
    environment = "pypi"
    sample_rate = 0.0 if is_local() else 1.0  # Don't send anything to Sentry for local testing
    sentry_init(environment=environment, sample_rate=sample_rate, **kwargs)

    def decorator(fn):
        @click.pass_context
        @wraps(fn)
        def wrapped(context: click.Context, *args, **kwargs):
            with configure_scope() as scope:
                # Set custom tags for Sentry alerts
                if org_slug := Config.get_org_slug():
                    scope.set_tag("org", org_slug)

                user = filter_dict({"username": getuser(), "email": Config.get_email()})
                scope.set_user(user)

                set_scope_context_os(scope)

                # Invoke the command
                return context.invoke(fn, *args, **kwargs)

        return wrapped

    return decorator



from typing import Optional

import click

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.commands.users.utils import get_or_prompt_service
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.tracked_command import TrackedCommand


@click.command(cls=TrackedCommand, name="update", short_help="Update an existing service")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.option(
    "--service-type",
    "_service_type",
    help="The service to update",
)
@click.option(
    "--external-id",
    "_external_id",
    help="The identifier for the service",
)
@click.option(
    "--label",
    "_label",
    help="The label for the service",
)
def services_update(
    options: GlobalOptions, _service_type: Optional[str], _external_id: Optional[str], _label: Optional[str]
) -> None:
    service = get_or_prompt_service(options.sym_api, _service_type, _external_id)
    options.sym_api.update_service(service.service_type, service.external_id, _label)
    cli_output.success(f"Successfully updated 1 service!")



from typing import Optional

import click

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.tracked_command import TrackedCommand
from sym.flow.cli.helpers.utils import get_or_prompt


@click.command(cls=TrackedCommand, name="set-role", short_help="Update a User's role")
@click.argument("email", required=True, type=str)
@click.option("--role", "role", help="The new role value", type=str)
@click.make_pass_decorator(GlobalOptions, ensure=True)
def set_role(options: GlobalOptions, email: str, role: Optional[str] = None) -> None:
    """For an existing Sym User, set their active role.

    \b
    - Guests may only approve, deny, or revoke access requests for flows that allow guest
    - Members may run any of your organization's flows, as well as do anything a guest can do.
    - Admins may manage other users and flows, as well as do anything a member can do.

    \b
    Example:
        `symflow users set-role user@symops.io --role admin`
    """

    api = options.sym_api
    user = api.get_user(email)
    role = get_or_prompt(role, "Which role?", ["admin", "member", "guest"])
    api.set_user_role(user.id, role)

    cli_output.success(f"Success! {email} now has the role '{role}'.")



from contextlib import contextmanager
from typing import Any, Optional

import yaml
from pydantic import BaseModel

from sym.flow.cli.helpers.utils import filter_dict
from sym.flow.cli.models import AuthToken, Organization

from .thread_safe_file import ThreadSafeFile


class ConfigFile(BaseModel):
    """A Pydantic model representing the contents of ~/.config/symflow/default/config.yml"""

    org: Optional[str] = None
    client_id: Optional[str] = None
    email: Optional[str] = None
    auth_token: Optional[AuthToken] = None

    @property
    def access_token(self) -> Optional[str]:
        if self.auth_token:
            return self.auth_token.access_token
        return None

    def deepget(self, key: str) -> Optional[Any]:
        """Tries to get a nested attribute from the config file (e.g. "auth_token.access_token")"""
        data = self
        for k in key.split("."):
            try:
                data = getattr(data, k)
            except AttributeError:
                return None

        return data


class Config:
    """A singleton class providing helper methods for accessing and modifying the symflow config file."""

    def __init__(self) -> None:
        self.file = ThreadSafeFile(path=(ThreadSafeFile.xdg_config_home() / "symflow/default/config.yml"))

        # Analogous to Python's built-in `with open(file)`, but with the locking mechanisms provided by ThreadSafeFile.
        with self.file.read() as f:
            # Initialize `self.config` with the config file from the file system.
            self.__load(f)

    @classmethod
    def reset(cls):
        """Replaces the singleton instance of this class with a new instance."""
        setattr(cls, "__instance", cls())

    @classmethod
    def instance(cls) -> "Config":
        """Returns a singleton instance of this class."""
        if not hasattr(Config, "__instance"):
            # If no instance exists yet, generate a new one
            Config.reset()

        # Return the existing instance
        return getattr(Config, "__instance")

    def __load(self, file_stream):
        """Loads the config file from the given file-stream into memory as a Pydantic model set as an attribute on the
        Singleton instance,"""
        self.config = ConfigFile(**(yaml.safe_load(stream=file_stream) or {}))

    def __flush(self, file_stream):
        """Writes the in-memory Pydantic model ConfigFile to the file-stream."""
        file_stream.seek(0)
        file_stream.truncate()
        yaml.safe_dump(filter_dict(self.config.dict()), stream=file_stream)

    @contextmanager
    def atomic(self):
        """A context manager that locks the config file for atomic updates."""

        # Analogous to Python's built-in `with open(file, 'r+')`,
        # but with the locking mechanisms provided by ThreadSafeFile.
        with self.file.update() as f:
            # Refresh the in-memory file from the file system.
            self.__load(f)
            yield
            # Write the new values to the system.
            self.__flush(f)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self.file)})"

    @classmethod
    def get_value(cls, key: str) -> Optional[str]:
        """Given a dot-separated string, attempts to get that attribute from the config.
        If the attribute doesn't exist, returns None.
        (e.g. key = "auth_token.access_token")
        """
        config = cls.instance().config
        return config.deepget(key)

    @classmethod
    def set_email(cls, email: str):
        config_manager = cls.instance()
        with config_manager.atomic():
            config = config_manager.config
            config.email = email

    @classmethod
    def get_email(cls) -> Optional[str]:
        return cls.instance().config.email

    @classmethod
    def set_auth_token(cls, auth_token: AuthToken):
        config_manager = cls.instance()
        with config_manager.atomic():
            config = config_manager.config
            config.auth_token = auth_token

    @classmethod
    def get_org(cls) -> Optional[Organization]:
        """Returns an Organization object containing the Org slug and Auth0 Client ID, if exists."""
        config = cls.instance().config
        if config.org:
            return Organization(slug=config.org, client_id=config.client_id)

        return None

    @classmethod
    def get_org_slug(cls) -> Optional[str]:
        """Returns the Org slug set in the config, if exists."""
        config = cls.instance().config
        return config.org

    @classmethod
    def set_org(cls, org: Organization):
        """Sets the org slug and Auth0 Client ID in the config file."""
        config_manager = cls.instance()
        with config_manager.atomic():
            config = config_manager.config
            config.org = org.slug
            config.client_id = org.client_id

    @classmethod
    def get_access_token(cls) -> Optional[str]:
        """Returns the Auth0 access token stored under `auth_token.access_token`"""
        return cls.instance().config.access_token

    @classmethod
    def is_logged_in(cls) -> bool:
        """Returns a boolean indicating if the user is logged in (i.e. has an auth_token set)."""
        return cls.instance().config.auth_token is not None

    @classmethod
    def logout(cls):
        """Logs out of symflow by removing the user's email and auth_token from the config file."""
        if not cls.is_logged_in():
            return

        config_manager = cls.instance()
        with config_manager.atomic():
            config = config_manager.config
            config.email = None
            config.auth_token = None

    @classmethod
    def store_login_config(cls, email: str, org: Organization, auth_token: AuthToken) -> str:
        """Updates the entire config file with the given values."""
        cls.set_org(org)
        cls.set_email(email)
        cls.set_auth_token(auth_token)
        return str(cls.instance().file)



from typing import Dict, List, Optional, Union

import click
from click import ClickException

from sym.flow.cli.helpers.constants import SYM_SUPPORT_EMAIL


class CliErrorMeta(type):
    """A metaclass that gives a unique exit code for each CliError class"""

    _exit_code_count = 2

    def __new__(cls, name, bases, attrs):
        cls._exit_code_count += 1
        klass = super().__new__(cls, name, bases, attrs)
        klass.exit_code = cls._exit_code_count
        return klass


class CliError(ClickException, metaclass=CliErrorMeta):
    """A superclass for all custom errors raised by symflow CLI."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

    def format_message(self):
        """Styles the outputted error messages of this class as bolded red text."""
        return click.style(self.message, fg="red", bold=True)


class CliErrorWithHint(CliError):
    """A superclass for custom errors raised by symflow CLI that have additional hint text."""

    def __init__(self, message, hints, **kwargs) -> None:
        self.hints = hints if isinstance(hints, list) else [hints]
        super().__init__(message, **kwargs)

    def __str__(self):
        return "\n\n".join(
            [
                self.message,
                *[f"Hint: {hint}" for hint in self.hints],
            ]
        )

    def format_message(self):
        """Styles the outputted error message as bolded red text, and hints as bolded cyan text."""
        msg = click.style(self.message, fg="red", bold=True)
        styled = [click.style(hint, fg="cyan", bold=True) for hint in self.hints]
        hint = "\n\n" + "\n\n".join([f"Hint: {hint}" for hint in styled])
        return f"{msg}{hint}"


class UnexpectedError(CliError):
    def __init__(self, ex: str) -> None:
        super().__init__(f"An unexpected error occurred: {ex}")


class SSLConnectionError(CliError):
    def __init__(self, ex: str) -> None:
        super().__init__(
            f"{ex}. This is usually caused by network monitoring/limiting software on the local computer. "
            f"Try disabling it. If the issue persists, please contact {SYM_SUPPORT_EMAIL}."
        )


class LoginError(CliErrorWithHint):
    def __init__(self, error_message: str, hint: str) -> None:
        super().__init__(error_message, hint)


class UnknownOrgError(CliError):
    def __init__(self, org_id: str) -> None:
        super().__init__(f"Unknown organization with ID: {org_id}")


class UnknownUserError(CliError):
    def __init__(self, email: str) -> None:
        super().__init__(f"Unknown user for email: {email}")


class UnknownBotError(CliError):
    def __init__(self, username: str) -> None:
        super().__init__(f"Unknown bot: {username}")


class UserAlreadyExists(CliError):
    def __init__(self, email: str) -> None:
        super().__init__(f"A user already exists with email: {email}")


class BotAlreadyExists(CliError):
    def __init__(self, username: str) -> None:
        super().__init__(f"A bot already exists with username: {username}")


class NotLoggedInError(CliErrorWithHint):
    def __init__(self) -> None:
        super().__init__(
            "You must be logged in to perform this action.",
            "Run `symflow login` to log in.",
        )


class MissingServiceError(CliErrorWithHint):
    def __init__(self, service_type: str) -> None:
        super().__init__(
            f"No service is registered for type {service_type}",
            f"You can create the service by declaring a `sym_integration` resource with `type={service_type}` in your Terraform configuration",
        )


class MissingIdentityValueError(CliErrorWithHint):
    def __init__(self, identifier: str, command: str = "users") -> None:
        super().__init__(
            "Identity value cannot be empty",
            f"If you want to delete the identity, run `symflow {command} delete-identity {identifier}`",
        )


class InvalidChoiceError(CliErrorWithHint):
    def __init__(self, value: str, valid_choices: List[str]) -> None:
        super().__init__(
            f"Invalid input: '{value}'",
            f"Try one of: {', '.join(valid_choices)}",
        )


class MissingChoicesError(CliError):
    def __init__(self) -> None:
        super().__init__("No choices were provided!")


class ReferencedObjectError(CliError):
    def __init__(self, references: Dict[str, List[str]]) -> None:
        counts = " and ".join(f"{len(refs)} {name}" for name, refs in references.items())

        super().__init__(f"Cannot perform delete because it is referenced by {counts}")


class InvalidExpiryError(CliErrorWithHint):
    def __init__(self, expiry: str) -> None:
        super().__init__(
            f"Invalid expiry input: {expiry}",
            f"Accepted values are a non-zero integer followed by s, m, d, or mo. e.g. 3d",
        )


class SymIdentityNotFound(CliError):
    def __init__(self, id: str) -> None:
        super().__init__(f"No Sym Identity found for user id {id}")


class SymAPIError(CliErrorWithHint):
    """Base exception for all API errors."""


class SymAPIUnauthorizedError(SymAPIError):
    def __init__(self, error_message: str) -> None:
        super().__init__(message="You are not authorized to perform this action.", hints=error_message)


class SymAPIRequestError(SymAPIError):
    def __init__(self, message: str, request_id: str, response_code: int) -> None:
        self.response_code = response_code
        self.request_id = request_id
        super().__init__("An API error occurred!", message)

    def format_message(self):
        """Formats the error message like a CliErrorWithHint, but appends an additional contact us message after the
        hint message.
        """
        formatted_message = super().format_message()
        contact_us_text = click.style(
            (
                f"\n\nPlease contact support and include your Request ID ({self.request_id})."
                f"\nhttps://docs.symops.com/docs/support"
            ),
            fg="white",
            bold=True,
        )
        return formatted_message + contact_us_text


class SymAPIAggregateError(SymAPIRequestError):
    def __init__(self, errors: Union[str, List[str]], request_id: str, response_code: int) -> None:
        self.errors = errors if isinstance(errors, list) else [errors]
        message = "\n\n".join([error for error in self.errors])
        super().__init__(message, request_id, response_code)


class SymAPIMissingEntityError(SymAPIRequestError):
    error_codes = [404]

    def __init__(self, response_code: int, request_id: str) -> None:
        super().__init__(f"Missing entity ({response_code}).", request_id, response_code)


class SymAPIUnknownError(SymAPIRequestError):
    """Errors returned by the Sym API that we do not know how to handle."""

    def __init__(self, response_code: int, request_id: str, message: Optional[str] = None) -> None:
        if message:
            super().__init__(f"An unexpected error occurred ({response_code}): {message}", request_id, response_code)
        else:
            super().__init__(f"An unknown error with status code {response_code}.", request_id, response_code)



import click

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.errors import SymAPIRequestError, UnknownBotError
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.tracked_command import TrackedCommand
from sym.flow.cli.models import ServiceType


@click.command(cls=TrackedCommand, name="delete", short_help="Delete a Sym Bot User")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.argument("username")
@click.option(
    "--force",
    is_flag=True,
    help="Delete the service immediately",
)
def bots_delete(options: GlobalOptions, username: str, force: bool) -> None:
    """
    Deletes a Bot User with the given username and revokes all their tokens.
    """

    # Prompt user to confirm
    if not force:
        click.confirm(
            "WARNING: This is a destructive action. \nAre you sure you want to continue?",
            abort=True,
            default=False,
        )

    payload = {
        "users": [
            {
                "identity": {
                    "service_type": ServiceType.SYM.type_name,
                    "matcher": {"username": username},
                },
            }
        ]
    }

    try:
        options.sym_api.delete_user(payload)
    except SymAPIRequestError as e:
        if e.response_code == 404:
            raise UnknownBotError(username)
    cli_output.success(f"Successfully deleted bot user {username}!")



import webbrowser

import click

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.helpers.constants import SYM_DOCUMENTATION_URL
from sym.flow.cli.helpers.tracked_command import TrackedCommand


@click.command(cls=TrackedCommand, short_help="Opens a link to Sym documentation")
def docs() -> None:
    """Opens a link to the Sym documentation website in a new window.

    Example:
        `symflow docs`
    """

    cli_output.info(f"Opening a link to Sym documentation: {SYM_DOCUMENTATION_URL}")
    webbrowser.open(SYM_DOCUMENTATION_URL)



import inspect
import logging
from dataclasses import dataclass
from typing import Optional, Sequence
from urllib.parse import urlparse

from sym.flow.cli.helpers.constants import DEFAULT_API_URL, DEFAULT_AUTH_URL


@dataclass
class GlobalOptions:
    api_url: str = DEFAULT_API_URL
    auth_url: str = DEFAULT_AUTH_URL
    access_token: Optional[str] = None
    debug: bool = False

    def dprint(self, *s: Sequence[str], **kwargs):
        """Prints debug messages if `self.debug` is True"""

        s = list(map(str, filter(None, s)))
        if (s or kwargs) and self.debug:
            message = " ".join(s)
            if kwargs:
                message += ": " + ",".join([f"{k}={v}" for k, v in kwargs.items()])
            mod = inspect.getmodule(inspect.stack()[1][0])
            logging.getLogger(mod.__name__ if mod else __name__).info(message)

    @property
    def sym_api(self):
        from sym.flow.cli.helpers.api import SymAPI

        return SymAPI(self.api_url, self.access_token)

    def set_access_token(self, sym_jwt):
        self.access_token = sym_jwt.strip()

    def set_api_url(self, url):
        url = url.rstrip("/")
        if not self._validate_url(url):
            raise ValueError(f"Specified SYM_API_URL '{url}' is not valid.")

        self.api_url = url

    def set_auth_url(self, url):
        url = url.rstrip("/")
        if not self._validate_url(url):
            raise ValueError(f"Specified SYM_AUTH_URL '{url}' is not valid.")

        self.auth_url = url

    def _validate_url(self, url):
        """Validates that the given URL:
        - Is `http` or `https`
        - Has a domain (e.g. api.symops.com)
        - Has no fragment (i.e. the part of the URL after `#`)
        - Has no query parameters
        - Has no param (i.e. the part between the path and the query)

        Details about the different parts of a URL:
        https://stackoverflow.com/questions/53992694/what-does-netloc-mean
        """
        try:
            parts = urlparse(url)
            if parts.scheme not in ("http", "https"):
                return False
            if not parts.netloc:
                return False
            if parts.fragment or parts.query or parts.params:
                return False
            return True
        except ValueError:
            return False

    def to_dict(self):
        return {
            "debug": self.debug,
            "api_url": self.api_url,
            "auth_url": self.auth_url,
        }



import click

import sym.flow.cli.helpers.output as cli_output
from sym.flow.cli.helpers.global_options import GlobalOptions
from sym.flow.cli.helpers.tracked_command import TrackedCommand


@click.command(cls=TrackedCommand, name="revoke", short_help="Revoke a Sym Token")
@click.make_pass_decorator(GlobalOptions, ensure=True)
@click.argument("jti")
def tokens_revoke(options: GlobalOptions, jti: str) -> None:
    """
    Revokes a token with the given identifier
    """
    options.sym_api.revoke_token(jti)
    cli_output.success(f"Successfully revoked token with identifier {jti}")



from typing import Any

from click import Command, Context

from sym.flow.cli.errors import SymAPIRequestError
from sym.flow.cli.helpers.constants import SegmentTrackStatus
from sym.flow.cli.helpers.global_options import GlobalOptions


class TrackedCommand(Command):
    """A custom Click Command that makes API calls to Queuer to track the command's invocation and success/fail statuses
    in Segment. Pass this in as the `cls` argument to the `@click.command` decorator for any symflow commands that
    should be tracked.

    For example:
    @click.command(cls=TrackedCommand, short_help="foo")
    def foo():
       ...
    """

    def invoke(self, ctx: Context) -> Any:
        # Pull the Global Options out of Click's magical Context.
        # This should always exist because we decorate all of our commands with
        # `@click.make_pass_decorator(GlobalOptions, ensure=True)`
        options: GlobalOptions
        if (options := ctx.find_object(GlobalOptions)) and options.sym_api.rest.access_token:

            # ctx.command_path returns the full command path starting from `symflow` (e.g. "symflow bots list").
            # We only want the subcommands, so strip the initial "symflow ".
            subcommand = ctx.command_path.replace("symflow ", "")

            options.sym_api.segment_track(subcommand, status=SegmentTrackStatus.INVOKED)
            try:
                result = super().invoke(ctx)
            except Exception as e:
                if isinstance(e, SymAPIRequestError):
                    # e.message for SymAPIRequestErrors are always "An API error occurred!", with the actual error
                    # is reflected in the hint message.
                    error_message = ", ".join(e.hints)
                else:
                    error_message = getattr(e, "message", str(e))
                options.sym_api.segment_track(subcommand, status=SegmentTrackStatus.ERROR, error=error_message)
                raise e

            options.sym_api.segment_track(subcommand, status=SegmentTrackStatus.SUCCESS)
            return result

        # If we don't have  Global Options or the user isn't logged in, then we can't track the command.
        # Just invoke the command normally.
        return super().invoke(ctx)



from enum import Enum

SentryDSN = "https://6a8722276347467b801e5aa152a48ed1@o373009.ingest.sentry.io/5462801"

DEFAULT_API_URL = "https://api.symops.com/api/v1"
DEFAULT_AUTH_URL = "https://auth.symops.com"

CLI_PACKAGE_NAME = "sym-flow-cli"

SYM_DOCUMENTATION_URL = "https://docs.symops.com/"
SYM_SUPPORT_EMAIL = "support@symops.com"


class SegmentTrackStatus(str, Enum):
    """Accepted statuses for the segment_track Sym API call."""

    INVOKED = "invoked"
    SUCCESS = "success"
    ERROR = "error"
